__all__ = ["Spotify", "SpotifyException"]

import base64
import os
import time
from pprint import pprint
from typing import Optional, Union

import requests
from dotenv import load_dotenv

from yutipy.exceptions import (
    AuthenticationException,
    InvalidResponseException,
    InvalidValueException,
    NetworkException,
    SpotifyException,
)
from yutipy.models import MusicInfo
from yutipy.utils.helpers import (
    are_strings_similar,
    guess_album_type,
    is_valid_string,
    separate_artists,
)
from yutipy.utils.logger import logger

load_dotenv()

SPOTIFY_CLIENT_ID = os.getenv("SPOTIFY_CLIENT_ID")
SPOTIFY_CLIENT_SECRET = os.getenv("SPOTIFY_CLIENT_SECRET")


class Spotify:
    """
    A class to interact with the Spotify API. It uses "Client Credentials" grant type (or flow).

    This class reads the ``SPOTIFY_CLIENT_ID`` and ``SPOTIFY_CLIENT_SECRET`` from environment variables or the ``.env`` file by default.
    Alternatively, you can manually provide these values when creating an object.
    """

    def __init__(
        self,
        client_id: str = SPOTIFY_CLIENT_ID,
        client_secret: str = SPOTIFY_CLIENT_SECRET,
    ) -> None:
        """
        Initializes the Spotify class (using Client Credentials grant type/flow) and sets up the session.

        Parameters
        ----------
        client_id : str, optional
            The Client ID for the Spotify API. Defaults to ``SPOTIFY_CLIENT_ID`` from environment variable or the ``.env`` file.
        client_secret : str, optional
            The Client secret for the Spotify API. Defaults to ``SPOTIFY_CLIENT_SECRET`` from environment variable or the ``.env`` file.
        """
        if not client_id or not client_secret:
            raise SpotifyException(
                "Failed to read `SPOTIFY_CLIENT_ID` and/or `SPOTIFY_CLIENT_SECRET` from environment variables. Client ID and Client Secret must be provided."
            )

        self.client_id = client_id
        self.client_secret = client_secret
        self._session = requests.Session()
        self.api_url = "https://api.spotify.com/v1"
        self.__header, self.__expires_in = self.__authenticate()
        self.__start_time = time.time()
        self._is_session_closed = False
        self.normalize_non_english = True
        self._translation_session = requests.Session()

    def __enter__(self):
        """Enters the runtime context related to this object."""
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        """Exits the runtime context related to this object."""
        self.close_session()

    def close_session(self) -> None:
        """Closes the current session(s)."""
        if not self.is_session_closed:
            self._session.close()
            self._translation_session.close()
            self._is_session_closed = True

    @property
    def is_session_closed(self) -> bool:
        """Checks if the session is closed."""
        return self._is_session_closed

    def __authenticate(self) -> tuple:
        """
        Authenticates with the Spotify API and returns the authorization header.

        Returns
        -------
        dict
            The authorization header.
        """
        try:
            token, expires_in = self.__get_spotify_token()
            return {"Authorization": f"Bearer {token}"}, expires_in
        except Exception as e:
            raise AuthenticationException(
                "Failed to authenticate with Spotify API"
            ) from e

    def __get_spotify_token(self) -> tuple:
        """
        Gets the Spotify API token.

        Returns
        -------
        str
            The Spotify API token.
        """
        auth_string = f"{self.client_id}:{self.client_secret}"
        auth_base64 = base64.b64encode(auth_string.encode("utf-8")).decode("utf-8")

        url = "https://accounts.spotify.com/api/token"
        headers = {
            "Authorization": f"Basic {auth_base64}",
            "Content-Type": "application/x-www-form-urlencoded",
        }
        data = {"grant_type": "client_credentials"}

        try:
            logger.info("Authenticating with Spotify API")
            response = self._session.post(
                url=url, headers=headers, data=data, timeout=30
            )
            logger.debug(f"Authentication response status code: {response.status_code}")
            response.raise_for_status()
        except requests.RequestException as e:
            logger.error(f"Network error during Spotify authentication: {e}")
            raise NetworkException(f"Network error occurred: {e}")

        try:
            response_json = response.json()
            return response_json.get("access_token"), response_json.get("expires_in")
        except (KeyError, ValueError) as e:
            raise InvalidResponseException(f"Invalid response received: {e}")

    def __refresh_token_if_expired(self):
        """Refreshes the token if it has expired."""
        if time.time() - self.__start_time >= self.__expires_in:
            self.__header, self.__expires_in = self.__authenticate()
            self.__start_time = time.time()

    def search(
        self,
        artist: str,
        song: str,
        limit: int = 10,
        normalize_non_english: bool = True,
    ) -> Optional[MusicInfo]:
        """
        Searches for a song by artist and title.

        Parameters
        ----------
        artist : str
            The name of the artist.
        song : str
            The title of the song.
        limit: int, optional
            The number of items to retrieve from API. ``limit >=1 and <= 50``. Default is ``10``.
        normalize_non_english : bool, optional
            Whether to normalize non-English characters for comparison. Default is ``True``.

        Returns
        -------
        Optional[MusicInfo_]
            The music information if found, otherwise None.
        """
        if not is_valid_string(artist) or not is_valid_string(song):
            raise InvalidValueException(
                "Artist and song names must be valid strings and can't be empty."
            )

        self.normalize_non_english = normalize_non_english

        music_info = None
        artist_ids = None
        queries = [
            f"?q=artist:{artist} track:{song}&type=track&limit={limit}",
            f"?q=artist:{artist} album:{song}&type=album&limit={limit}",
        ]

        for query in queries:
            if music_info:
                return music_info

            self.__refresh_token_if_expired()

            query_url = f"{self.api_url}/search{query}"

            logger.info(
                f"Searching Spotify for `artist='{artist}'` and `song='{song}'`"
            )
            logger.debug(f"Query URL: {query_url}")

            try:
                response = self._session.get(
                    query_url, headers=self.__header, timeout=30
                )
                response.raise_for_status()
            except requests.RequestException as e:
                raise NetworkException(f"Network error occurred: {e}")

            if response.status_code != 200:
                raise SpotifyException(f"Failed to search for music: {response.json()}")

            artist_ids = artist_ids if artist_ids else self._get_artists_ids(artist)
            music_info = self._find_music_info(
                artist, song, response.json(), artist_ids
            )

        return music_info

    def search_advanced(
        self,
        artist: str,
        song: str,
        isrc: str = None,
        upc: str = None,
        limit: int = 1,
        normalize_non_english: bool = True,
    ) -> Optional[MusicInfo]:
        """
        Searches for a song by artist, title, ISRC, or UPC.

        Parameters
        ----------
        artist : str
            The name of the artist.
        song : str
            The title of the song.
        isrc : str, optional
            The ISRC of the track.
        upc : str, optional
            The UPC of the album.
        limit: int, optional
            The number of items to retrieve from API. ``limit >=1 and <= 50``. Default is ``1``.
        normalize_non_english : bool, optional
            Whether to normalize non-English characters for comparison. Default is ``True``.

        Returns
        -------
        Optional[MusicInfo_]
            The music information if found, otherwise None.
        """
        if not is_valid_string(artist) or not is_valid_string(song):
            raise InvalidValueException(
                "Artist and song names must be valid strings and can't be empty."
            )

        self.normalize_non_english = normalize_non_english

        self.__refresh_token_if_expired()

        if isrc:
            query = f"?q={artist} {song} isrc:{isrc}&type=track&limit={limit}"
        elif upc:
            query = f"?q={artist} {song} upc:{upc}&type=album&limit={limit}"
        else:
            raise InvalidValueException("ISRC or UPC must be provided.")

        query_url = f"{self.api_url}/search{query}"
        try:
            response = self._session.get(query_url, headers=self.__header, timeout=30)
            response.raise_for_status()
        except requests.RequestException as e:
            raise NetworkException(f"Network error occurred: {e}")

        if response.status_code != 200:
            raise SpotifyException(
                f"Failed to search music with ISRC/UPC: {response.json()}"
            )

        artist_ids = self._get_artists_ids(artist)
        return self._find_music_info(artist, song, response.json(), artist_ids)

    def _get_artists_ids(self, artist: str) -> Union[list, None]:
        """
        Retrieves the IDs of the artists.

        Parameters
        ----------
        artist : str
            The name of the artist.

        Returns
        -------
        Union[list, None]
            A list of artist IDs or None if not found.
        """
        artist_ids = []
        for name in separate_artists(artist):
            query_url = f"{self.api_url}/search?q={name}&type=artist&limit=5"
            try:
                response = self._session.get(
                    query_url, headers=self.__header, timeout=30
                )
                response.raise_for_status()
            except requests.RequestException as e:
                raise NetworkException(f"Network error occurred: {e}")

            if response.status_code != 200:
                return None

            artist_ids.extend(
                artist["id"] for artist in response.json()["artists"]["items"]
            )
        return artist_ids

    def _find_music_info(
        self, artist: str, song: str, response_json: dict, artist_ids: list
    ) -> Optional[MusicInfo]:
        """
        Finds the music information from the search results.

        Parameters
        ----------
        artist : str
            The name of the artist.
        song : str
            The title of the song.
        response_json : dict
            The JSON response from the API.
        artist_ids : list
            A list of artist IDs.

        Returns
        -------
        Optional[MusicInfo]
            The music information if found, otherwise None.
        """
        try:
            for track in response_json["tracks"]["items"]:
                music_info = self._find_track(song, artist, track, artist_ids)
                if music_info:
                    return music_info
        except KeyError:
            pass

        try:
            for album in response_json["albums"]["items"]:
                music_info = self._find_album(song, artist, album, artist_ids)
                if music_info:
                    return music_info
        except KeyError:
            pass

        logger.warning(
            f"No matching results found for artist='{artist}' and song='{song}'"
        )
        return None

    def _find_track(
        self, song: str, artist: str, track: dict, artist_ids: list
    ) -> Optional[MusicInfo]:
        """
        Finds the track information from the search results.

        Parameters
        ----------
        song : str
            The title of the song.
        artist : str
            The name of the artist.
        track : dict
            A single track from the search results.
        artist_ids : list
            A list of artist IDs.

        Returns
        -------
        Optional[MusicInfo]
            The music information if found, otherwise None.
        """
        if not are_strings_similar(
            track["name"],
            song,
            use_translation=self.normalize_non_english,
            translation_session=self._translation_session,
        ):
            return None

        artists_name = [x["name"] for x in track["artists"]]
        matching_artists = [
            x["name"]
            for x in track["artists"]
            if are_strings_similar(
                x["name"],
                artist,
                use_translation=self.normalize_non_english,
                translation_session=self._translation_session,
            )
            or x["id"] in artist_ids
        ]

        if matching_artists:
            return MusicInfo(
                album_art=track["album"]["images"][0]["url"],
                album_title=track["album"]["name"],
                album_type=track["album"]["album_type"],
                artists=", ".join(artists_name),
                genre=None,
                id=track["id"],
                isrc=track.get("external_ids").get("isrc"),
                lyrics=None,
                release_date=track["album"]["release_date"],
                tempo=None,
                title=track["name"],
                type="track",
                upc=None,
                url=track["external_urls"]["spotify"],
            )

        return None

    def _find_album(
        self, song: str, artist: str, album: dict, artist_ids: list
    ) -> Optional[MusicInfo]:
        """
        Finds the album information from the search results.

        Parameters
        ----------
        song : str
            The title of the song.
        artist : str
            The name of the artist.
        album : dict
            A single album from the search results.
        artist_ids : list
            A list of artist IDs.

        Returns
        -------
        Optional[MusicInfo]
            The music information if found, otherwise None.
        """
        if not are_strings_similar(
            album["name"],
            song,
            use_translation=self.normalize_non_english,
            translation_session=self._translation_session,
        ):
            return None

        artists_name = [x["name"] for x in album["artists"]]
        matching_artists = [
            x["name"]
            for x in album["artists"]
            if are_strings_similar(
                x["name"],
                artist,
                use_translation=self.normalize_non_english,
                translation_session=self._translation_session,
            )
            or x["id"] in artist_ids
        ]

        if matching_artists:
            guess = guess_album_type(album.get("total_tracks", 1))
            guessed_right = are_strings_similar(
                album.get("album_type", "x"), guess, use_translation=False
            )

            return MusicInfo(
                album_art=album["images"][0]["url"],
                album_title=album["name"],
                album_type=album.get("album_type") if guessed_right else guess,
                artists=", ".join(artists_name),
                genre=None,
                id=album["id"],
                isrc=None,
                lyrics=None,
                release_date=album["release_date"],
                tempo=None,
                title=album["name"],
                type=album.get("type"),
                upc=None,
                url=album["external_urls"]["spotify"],
            )

        return None


if __name__ == "__main__":
    import logging

    from yutipy.utils.logger import enable_logging

    enable_logging(level=logging.DEBUG)
    spotify = Spotify(SPOTIFY_CLIENT_ID, SPOTIFY_CLIENT_SECRET)

    try:
        artist_name = input("Artist Name: ")
        song_name = input("Song Name: ")
        pprint(spotify.search(artist_name, song_name))
    finally:
        spotify.close_session()



import os
import webbrowser
from dotenv import load_dotenv, set_key


def run_config_wizard():
    """Interactive configuration wizard for setting up API keys."""
    print("Welcome to the yutipy Configuration Wizard!")
    print("This wizard will help you set up your API keys for various services.\n")

    # Load existing .env file if it exists
    env_file = ".env"
    load_dotenv(env_file)

    # List of required environment variables and their instructions
    required_vars = {
        "SPOTIFY_CLIENT_ID": {
            "description": "Spotify Client ID",
            "url": "https://developer.spotify.com/dashboard",
            "instructions": """
1. Go to your Spotify Developer Dashboard: https://developer.spotify.com/dashboard
2. Create a new app and fill in the required details.
3. Copy the "Client ID" and "Client Secret" from the app's settings.
4. Paste them here when prompted.
            """,
        },
        "SPOTIFY_CLIENT_SECRET": {
            "description": "Spotify Client Secret",
            "url": "https://developer.spotify.com/dashboard",
            "instructions": "See the steps above for Spotify Client ID.",
        },
        "KKBOX_CLIENT_ID": {
            "description": "KKBox Client ID",
            "url": "https://developer.kkbox.com/",
            "instructions": """
1. Go to the KKBOX Developer Portal: https://developer.kkbox.com/
2. Log in and create a new application.
3. Copy the "Client ID" and "Client Secret" from the app's settings.
4. Paste them here when prompted.
            """,
        },
        "KKBOX_CLIENT_SECRET": {
            "description": "KKBox Client Secret",
            "url": "https://developer.kkbox.com/",
            "instructions": "See the steps above for KKBox Client ID.",
        },
    }

    # Track whether the browser has already been opened for a service
    browser_opened = set()

    # Prompt the user for each variable
    for var, details in required_vars.items():
        current_value = os.getenv(var)
        if current_value:
            print(f"{details['description']} is already set.")
            continue

        print(f"\n{details['description']} is missing.")
        print(details["instructions"])

        # Check if the browser has already been opened for this service
        if details["url"] not in browser_opened:
            open_browser = (
                input(
                    f"Do you want to open the website to get your {details['description']}? (y/N): "
                )
                .strip()
                .lower()
            )
            if open_browser == "y":
                webbrowser.open(details["url"])
                print(f"The website has been opened in your browser: {details['url']}")
                browser_opened.add(details["url"])  # Mark this URL as opened

        # Prompt the user to enter the value
        new_value = input(f"Enter your {details['description']}: ").strip()
        if new_value:
            set_key(env_file, var, new_value)
            print(f"{details['description']} has been saved to the .env file.")

    print("\nConfiguration complete! Your API keys have been saved to the .env file.")


if __name__ == "__main__":
    run_config_wizard()



__all__ = [
    "AuthenticationException",
    "InvalidResponseException",
    "InvalidValueException",
    "NetworkException",
    "YutipyException",
]


# Base Exception
class YutipyException(Exception):
    """Base class for exceptions in the Yutipy package."""

    pass


# Service Exceptions
class DeezerException(YutipyException):
    """Exception raised for errors related to the Deezer API."""

    pass


class ItunesException(YutipyException):
    """Exception raised for errors related to the iTunes API."""

    pass


class KKBoxException(YutipyException):
    """Exception raised for erros related to the KKBOX Open API."""

    pass


class MusicYTException(YutipyException):
    """Exception raised for errors related to the YouTube Music API."""

    pass


class SpotifyException(YutipyException):
    """Exception raised for errors related to the Spotify API."""

    pass


# Generic Exceptions
class AuthenticationException(YutipyException):
    """Exception raised for authentication errors."""

    pass


class InvalidResponseException(YutipyException):
    """Exception raised for invalid responses from APIs."""

    pass


class InvalidValueException(YutipyException):
    """Exception raised for invalid values."""

    pass


class NetworkException(YutipyException):
    """Exception raised for network-related errors."""

    pass



import pytest
from yutipy.utils import are_strings_similar, separate_artists


def test_are_strings_similar():
    assert are_strings_similar("Hello World", "Hello World") is True
    assert are_strings_similar("Hello World", "Hello") is True


def test_are_strings_similar_translation():
    assert are_strings_similar("ポーター", "Porter") is True


def test_separate_artists():
    assert separate_artists("Artist A & Artist B") == ["Artist A", "Artist B"]
    assert separate_artists("Artist A ft. Artist B") == ["Artist A", "Artist B"]
    assert separate_artists("Artist A") == ["Artist A"]
    assert separate_artists("Artist A and Artist B") == ["Artist A", "Artist B"]
    assert separate_artists("Artist A / Artist B") == ["Artist A", "Artist B"]
    assert separate_artists("Artist A, Artist B", custom_separator=",") == [
        "Artist A",
        "Artist B",
    ]



from concurrent.futures import ThreadPoolExecutor, as_completed
from pprint import pprint
from typing import Optional

from yutipy.deezer import Deezer
from yutipy.exceptions import InvalidValueException, KKBoxException, SpotifyException
from yutipy.itunes import Itunes
from yutipy.kkbox import KKBox
from yutipy.models import MusicInfo, MusicInfos
from yutipy.musicyt import MusicYT
from yutipy.spotify import Spotify
from yutipy.utils.helpers import is_valid_string
from yutipy.utils.logger import logger


class YutipyMusic:
    """A class that can be used to retrieve music information from all music platforms available in ``yutipy``.

    This is useful when you want to get music information (especially streaming link) from all available platforms.
    Instead of calling each service separately, you can use this class to get the information from all services at once.
    """

    def __init__(self) -> None:
        """Initializes the YutipyMusic class."""
        self.music_info = MusicInfos()
        self.normalize_non_english = True
        self.album_art_priority = ["deezer", "ytmusic", "itunes"]
        self.services = {
            "deezer": Deezer(),
            "itunes": Itunes(),
            "ytmusic": MusicYT(),
        }

        try:
            self.services["kkbox"] = KKBox()
        except (KKBoxException) as e:
            logger.warning(
                f"{self.__class__.__name__}: Skipping KKBox due to KKBoxException: {e}"
            )
        else:
            idx = self.album_art_priority.index("ytmusic")
            self.album_art_priority.insert(idx, "kkbox")

        try:
            self.services["spotify"] = Spotify()
        except (SpotifyException) as e:
            logger.warning(
                f"{self.__class__.__name__}: Skipping Spotify due to SpotifyException: {e}"
            )
        else:
            idx = self.album_art_priority.index("ytmusic")
            self.album_art_priority.insert(idx, "spotify")

    def __enter__(self) -> "YutipyMusic":
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback) -> None:
        self.close_sessions()

    def search(
        self,
        artist: str,
        song: str,
        limit: int = 5,
        normalize_non_english: bool = True,
    ) -> Optional[MusicInfos]:
        """
        Searches for a song by artist and title.

        Parameters
        ----------
        artist : str
            The name of the artist.
        song : str
            The title of the song.
        limit: int, optional
            The number of items to retrieve from all APIs. ``limit >=1 and <= 50``. Default is ``5``.
        normalize_non_english : bool, optional
            Whether to normalize non-English characters for comparison. Default is ``True``.

        Returns
        -------
        Optional[MusicInfos_]
            The music information if found, otherwise None.
        """
        if not is_valid_string(artist) or not is_valid_string(song):
            raise InvalidValueException(
                "Artist and song names must be valid strings and can't be empty."
            )

        self.normalize_non_english = normalize_non_english

        logger.info(
            f"Searching all platforms for `artist='{artist}'` and `song='{song}'`"
        )

        with ThreadPoolExecutor() as executor:
            futures = {
                executor.submit(
                    service.search,
                    artist=artist,
                    song=song,
                    limit=limit,
                    normalize_non_english=self.normalize_non_english,
                ): name
                for name, service in self.services.items()
            }

            for future in as_completed(futures):
                service_name = futures[future]
                try:
                    result = future.result()
                    self._combine_results(result, service_name)
                except Exception as e:
                    logger.error(
                        f"Error occurred while searching with {service_name}: {e}"
                    )

        if len(self.music_info.url) == 0:
            logger.warning(
                f"No matching results found across all platforms for artist='{artist}' and song='{song}'"
            )
            return None

        return self.music_info

    def _combine_results(self, result: Optional[MusicInfo], service_name: str) -> None:
        """
        Combines the results from different services.

        Parameters
        ----------
        result : Optional[MusicInfo]
            The music information from a service.
        service_name : str
            The name of the streaming service.
        """
        if not result:
            return

        attributes = [
            "album_title",
            "album_type",
            "artists",
            "genre",
            "isrc",
            "lyrics",
            "release_date",
            "tempo",
            "title",
            "type",
            "upc",
        ]

        for attr in attributes:
            if getattr(result, attr) and (
                not getattr(self.music_info, attr)
                or (attr in ["genre", "album_type"] and service_name == "itunes")
            ):
                setattr(self.music_info, attr, getattr(result, attr))

        if result.album_art:
            current_priority = self.album_art_priority.index(service_name)
            existing_priority = (
                self.album_art_priority.index(self.music_info.album_art_source)
                if self.music_info.album_art_source
                else len(self.album_art_priority)
            )
            if current_priority < existing_priority:
                self.music_info.album_art = result.album_art
                self.music_info.album_art_source = service_name

        self.music_info.id[service_name] = result.id
        self.music_info.url[service_name] = result.url

    def close_sessions(self) -> None:
        """Closes the sessions for all services."""
        for service in self.services.values():
            if hasattr(service, "close_session"):
                service.close_session()


if __name__ == "__main__":
    import logging
    from yutipy.utils.logger import enable_logging

    enable_logging(level=logging.DEBUG)
    yutipy_music = YutipyMusic()

    artist_name = input("Artist Name: ")
    song_name = input("Song Name: ")

    pprint(yutipy_music.search(artist_name, song_name))
    yutipy_music.close_sessions()



from .deezer import Deezer
from .itunes import Itunes
from .kkbox import KKBox
from .musicyt import MusicYT
from .spotify import Spotify
from .utils import disable_logging, enable_logging
from .yutipy_music import YutipyMusic

__all__ = [
    "Deezer",
    "Itunes",
    "KKBox",
    "MusicYT",
    "Spotify",
    "YutipyMusic",
    "enable_logging",
    "disable_logging",
]



import pytest
from pytest import raises

from yutipy.itunes import Itunes
from yutipy.models import MusicInfo
from yutipy.exceptions import InvalidValueException


@pytest.fixture
def itunes():
    return Itunes()


def test_search_valid(itunes):
    artist = "Adele"
    song = "Hello"
    result = itunes.search(artist, song)
    assert result is not None
    assert isinstance(result, MusicInfo)
    assert "Hello" in result.title
    assert "Adele" in result.artists


def test_search_invalid(itunes):
    artist = "Nonexistent Artist"
    song = "Nonexistent Song"
    result = itunes.search(artist, song)
    assert result is None


def test_search_empty_artist(itunes):
    artist = ""
    song = "Hello"

    with raises(InvalidValueException):
        itunes.search(artist, song)


def test_search_empty_song(itunes):
    artist = "Adele"
    song = ""

    with raises(InvalidValueException):
        itunes.search(artist, song)


def test_close_session(itunes):
    itunes.close_session()
    assert itunes.is_session_closed



import pytest

from yutipy.deezer import Deezer
from yutipy.models import MusicInfo


@pytest.fixture
def deezer():
    return Deezer()


def test_search_valid(deezer):
    result = deezer.search("Adele", "Hello")
    assert result is not None
    assert isinstance(result, MusicInfo)
    assert result.title == "Hello"


def test_search_invalid(deezer):
    result = deezer.search("Nonexistent Artist", "Nonexistent Song")
    assert result is None


def test_search_translation_matching(deezer):
    result = deezer.search("Porter Robinson", "Shelter", normalize_non_english=False)
    assert result.id != 1355346522
    result = deezer.search("Porter Robinson", "Shelter", normalize_non_english=True)
    assert result.id == 1355346522 or 2404408865


def test_get_upc_isrc_track(deezer):
    track_id = 781592622
    result = deezer._get_upc_isrc(track_id, "track")
    assert result is not None
    assert "isrc" in result
    assert "release_date" in result


def test_get_upc_isrc_album(deezer):
    album_id = 115888392
    result = deezer._get_upc_isrc(album_id, "album")
    assert result is not None
    assert "upc" in result
    assert "release_date" in result


def test_search_no_results(deezer):
    result = deezer.search("Adele", "Nonexistent Song")
    assert result is None


def test_close_session(deezer):
    deezer.close_session()
    assert deezer.is_session_closed



from .helpers import (
    are_strings_similar,
    guess_album_type,
    is_valid_string,
    separate_artists,
)
from .logger import disable_logging, enable_logging

__all__ = [
    "guess_album_type",
    "are_strings_similar",
    "is_valid_string",
    "separate_artists",
    "enable_logging",
    "disable_logging"
]



__all__ = ["Deezer", "DeezerException"]

from pprint import pprint
from typing import Dict, List, Optional

import requests

from yutipy.exceptions import (
    DeezerException,
    InvalidResponseException,
    InvalidValueException,
    NetworkException,
)
from yutipy.models import MusicInfo
from yutipy.utils.helpers import are_strings_similar, is_valid_string
from yutipy.utils.logger import logger


class Deezer:
    """A class to interact with the Deezer API."""

    def __init__(self) -> None:
        """Initializes the Deezer class and sets up the session."""
        self._session = requests.Session()
        self.api_url = "https://api.deezer.com"
        self._is_session_closed = False
        self.normalize_non_english = True
        self._translation_session = requests.Session()

    def __enter__(self) -> "Deezer":
        """Enters the runtime context related to this object."""
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback) -> None:
        """Exits the runtime context related to this object."""
        self.close_session()

    def close_session(self) -> None:
        """Closes the current session."""
        if not self.is_session_closed:
            self._session.close()
            self._translation_session.close()
            self._is_session_closed = True

    @property
    def is_session_closed(self) -> bool:
        """Checks if the session is closed."""
        return self._is_session_closed

    def search(
        self,
        artist: str,
        song: str,
        limit: int = 10,
        normalize_non_english: bool = True,
    ) -> Optional[MusicInfo]:
        """
        Searches for a song by artist and title.

        Parameters
        ----------
        artist : str
            The name of the artist.
        song : str
            The title of the song.
        limit: int, optional
            The number of items to retrieve from API. ``limit >=1 and <= 50``. Default is ``10``.
        normalize_non_english : bool, optional
            Whether to normalize non-English characters for comparison. Default is ``True``.


        Returns
        -------
        Optional[MusicInfo_]
            The music information if found, otherwise None.
        """
        if not is_valid_string(artist) or not is_valid_string(song):
            raise InvalidValueException(
                "Artist and song names must be valid strings and can't be empty."
            )

        self.normalize_non_english = normalize_non_english

        search_types = ["track", "album"]

        for search_type in search_types:
            endpoint = f"{self.api_url}/search/{search_type}"
            query = f'?q=artist:"{artist}" {search_type}:"{song}"&limit={limit}'
            query_url = endpoint + query

            try:
                logger.info(
                    f'Searching music info for `artist="{artist}"` and `song="{song}"`'
                )
                logger.debug(f"Query URL: {query_url}")
                response = self._session.get(query_url, timeout=30)
                logger.debug(f"Response status code: {response.status_code}")
                response.raise_for_status()
            except requests.RequestException as e:
                logger.error(f"Network error while fetching music info: {e}")
                raise NetworkException(f"Network error occurred: {e}")
            except Exception as e:
                logger.exception(f"Unexpected error while searching Deezer: {e}")
                raise DeezerException(f"An error occurred while searching Deezer: {e}")

            try:
                logger.debug(f"Parsing response JSON: {response.json()}")
                result = response.json()["data"]
            except (IndexError, KeyError, ValueError) as e:
                logger.error(f"Invalid response structure from Deezer: {e}")
                raise InvalidResponseException(f"Invalid response received: {e}")

            music_info = self._parse_results(artist, song, result)
            if music_info:
                return music_info

        logger.warning(
            f"No matching results found for artist='{artist}' and song='{song}'"
        )
        return None

    def _get_upc_isrc(self, music_id: int, music_type: str) -> Optional[Dict]:
        """
        Retrieves UPC and ISRC information for a given music ID and type.

        Parameters
        ----------
        music_id : int
            The ID of the music.
        music_type : str
            The type of the music (track or album).

        Returns
        -------
        Optional[Dict]
            A dictionary containing UPC and ISRC information.
        """
        if music_type == "track":
            return self._get_track_info(music_id)
        elif music_type == "album":
            return self._get_album_info(music_id)
        else:
            raise DeezerException(f"Invalid music type: {music_type}")

    def _get_track_info(self, track_id: int) -> Optional[Dict]:
        """
        Retrieves track information for a given track ID.

        Parameters
        ----------
        music_id : int
            The ID of the track.

        Returns
        -------
        Optional[Dict]
            A dictionary containing track information.
        """
        query_url = f"{self.api_url}/track/{track_id}"
        try:
            logger.info(f"Fetching track info for track_id: {track_id}")
            logger.debug(f"Query URL: {query_url}")
            response = self._session.get(query_url, timeout=30)
            logger.debug(f"Response status code: {response.status_code}")
            response.raise_for_status()
        except requests.RequestException as e:
            logger.error(f"Error fetching track info: {e}")
            raise NetworkException(f"Network error occurred: {e}")
        except Exception as e:
            logger.error(f"Error fetching track info: {e}")
            raise DeezerException(f"An error occurred while fetching track info: {e}")

        try:
            logger.debug(f"Response JSON: {response.json()}")
            result = response.json()
        except ValueError as e:
            logger.error(f"Invalid response received from Deezer: {e}")
            raise InvalidResponseException(f"Invalid response received: {e}")

        return {
            "isrc": result.get("isrc"),
            "release_date": result.get("release_date"),
            "tempo": result.get("bpm"),
        }

    def _get_album_info(self, album_id: int) -> Optional[Dict]:
        """
        Retrieves album information for a given album ID.

        Parameters
        ----------
        music_id : int
            The ID of the album.

        Returns
        -------
        Optional[Dict]
            A dictionary containing album information.
        """
        query_url = f"{self.api_url}/album/{album_id}"
        try:
            logger.info(f"Fetching album info for album_id: {album_id}")
            logger.debug(f"Query URL: {query_url}")
            response = self._session.get(query_url, timeout=30)
            logger.info(f"Response status code: {response.status_code}")
            response.raise_for_status()
        except requests.RequestException as e:
            logger.error(f"Error fetching album info: {e}")
            raise NetworkException(f"Network error occurred: {e}")
        except Exception as e:
            logger.error(f"Error fetching album info: {e}")
            raise DeezerException(f"An error occurred while fetching album info: {e}")

        try:
            logger.debug(f"Response JSON: {response.json()}")
            result = response.json()
        except ValueError as e:
            logger.error(f"Invalid response received from Deezer: {e}")
            raise InvalidResponseException(f"Invalid response received: {e}")

        return {
            "genre": (
                result["genres"]["data"][0]["name"]
                if result["genres"]["data"]
                else None
            ),
            "release_date": result.get("release_date"),
            "upc": result.get("upc"),
        }

    def _parse_results(
        self, artist: str, song: str, results: List[Dict]
    ) -> Optional[MusicInfo]:
        """
        Parses the search results to find a matching song.

        Parameters
        ----------
        artist : str
            The name of the artist.
        song : str
            The title of the song.
        results : List[Dict]
            The search results from the API.

        Returns
        -------
        Optional[MusicInfo]
            The music information if a match is found, otherwise None.
        """
        for result in results:
            if not (
                are_strings_similar(
                    result["title"],
                    song,
                    use_translation=self.normalize_non_english,
                    translation_session=self._translation_session,
                )
                and are_strings_similar(
                    result["artist"]["name"],
                    artist,
                    use_translation=self.normalize_non_english,
                    translation_session=self._translation_session,
                )
            ):
                continue

            return self._extract_music_info(result)

        return None

    def _extract_music_info(self, result: Dict) -> MusicInfo:
        """
        Extracts music information from a search result.

        Parameters
        ----------
        result : Dict
            A single search result from the API.

        Returns
        -------
        MusicInfo
            The extracted music information.
        """
        music_type = result["type"]
        music_info = MusicInfo(
            album_art=(
                result["album"]["cover_xl"]
                if music_type == "track"
                else result["cover_xl"]
            ),
            album_title=(
                result["album"]["title"] if music_type == "track" else result["title"]
            ),
            album_type=result.get("record_type", music_type.replace("track", "single")),
            artists=result["artist"]["name"],
            genre=None,
            id=result["id"],
            isrc=None,
            lyrics=None,
            release_date=None,
            tempo=None,
            title=result["title"],
            type=music_type,
            upc=None,
            url=result["link"],
        )

        if music_type == "track":
            track_info = self._get_upc_isrc(result["id"], music_type)
            music_info.isrc = track_info.get("isrc")
            music_info.release_date = track_info.get("release_date")
            music_info.tempo = track_info.get("tempo")
        else:
            album_info = self._get_upc_isrc(result["id"], music_type)
            music_info.upc = album_info.get("upc")
            music_info.release_date = album_info.get("release_date")
            music_info.genre = album_info.get("genre")

        return music_info


if __name__ == "__main__":
    import logging
    from yutipy.utils.logger import enable_logging

    enable_logging(level=logging.DEBUG)
    deezer = Deezer()
    try:
        artist_name = input("Artist Name: ")
        song_name = input("Song Name: ")
        pprint(deezer.search(artist_name, song_name))
    finally:
        deezer.close_session()



import pytest
from yutipy.yutipy_music import YutipyMusic
from yutipy.models import MusicInfos
from yutipy.exceptions import InvalidValueException


@pytest.fixture
def yutipy_music():
    return YutipyMusic()


def test_search_valid(yutipy_music):
    artist = "Adele"
    song = "Hello"
    result = yutipy_music.search(artist, song)
    assert result is not None
    assert isinstance(result, MusicInfos)
    assert "Hello" in result.title
    assert "Adele" in result.artists


def test_search_invalid(yutipy_music):
    artist = "Nonexistent Artist"
    song = "Nonexistent Song"
    result = yutipy_music.search(artist, song)
    assert result is None


def test_search_empty_artist(yutipy_music):
    artist = ""
    song = "Hello"

    with pytest.raises(InvalidValueException):
        yutipy_music.search(artist, song)


def test_search_empty_song(yutipy_music):
    artist = "Adele"
    song = ""

    with pytest.raises(InvalidValueException):
        yutipy_music.search(artist, song)


def test_close_sessions(yutipy_music):
    yutipy_music.close_sessions()
    for service in yutipy_music.services.values():
        assert service.is_session_closed



import pytest

from yutipy.exceptions import SpotifyException
from yutipy.models import MusicInfo
from yutipy.spotify import Spotify


@pytest.fixture(scope="module")
def spotify():
    try:
        return Spotify()
    except SpotifyException:
        pytest.skip("Spotify credentials not found")


def test_search(spotify):
    artist = "Adele"
    song = "Hello"
    result = spotify.search(artist, song)
    assert result is not None
    assert isinstance(result, MusicInfo)
    assert result.title == song
    assert artist in result.artists


def test_search_advanced_with_isrc(spotify):
    artist = "Adele"
    song = "Hello"
    isrc = "GBBKS1500214"
    result = spotify.search_advanced(artist, song, isrc=isrc)
    assert result is not None
    assert result.isrc == isrc


def test_search_advanced_with_upc(spotify):
    artist = "Miles Davis"
    album = "Kind Of Blue (Legacy Edition)"
    upc = "888880696069"
    result = spotify.search_advanced(artist, album, upc=upc)
    print(result)
    assert result is not None


def test_get_artists_ids(spotify):
    artist = "Adele"
    artist_ids = spotify._get_artists_ids(artist)
    assert isinstance(artist_ids, list)
    assert len(artist_ids) > 0


def test_close_session(spotify):
    spotify.close_session()
    assert spotify.is_session_closed



import pytest
from pytest import raises

from yutipy.exceptions import KKBoxException, InvalidValueException
from yutipy.models import MusicInfo
from yutipy.kkbox import KKBox


@pytest.fixture(scope="module")
def kkbox():
    try:
        return KKBox()
    except KKBoxException:
        pytest.skip("KKBOX credentials not found")


def test_search(kkbox):
    artist = "Porter Robinson"
    song = "Shelter"
    result = kkbox.search(artist, song)
    assert result is not None
    assert isinstance(result, MusicInfo)
    assert result.title == song
    assert artist in result.artists


def test_get_html_widget(kkbox):
    html_widget = kkbox.get_html_widget(id="8rceGrek59bDS0HmQH", content_type="song")
    assert html_widget is not None
    assert isinstance(html_widget, str)

    with raises(InvalidValueException):
        kkbox.get_html_widget(id="8rceGrek59bDS0HmQH", content_type="track")

    with raises(InvalidValueException):
        kkbox.get_html_widget(id="8rceGrek59bDS0HmQH", content_type="song", territory="US")

    with raises(InvalidValueException):
        kkbox.get_html_widget(id="8rceGrek59bDS0HmQH", content_type="song", widget_lang="JP")


def test_close_session(kkbox):
    kkbox.close_session()
    assert kkbox.is_session_closed



import pytest
from pytest import raises
from yutipy.musicyt import MusicYT
from yutipy.models import MusicInfo
from yutipy.exceptions import InvalidValueException


@pytest.fixture
def music_yt():
    return MusicYT()


def test_search_valid(music_yt):
    artist = "Adele"
    song = "Hello (Official Music Video)"
    result = music_yt.search(artist, song)
    assert result is not None
    assert isinstance(result, MusicInfo)
    assert artist in result.artists
    assert result.title == song


def test_search_invalid(music_yt):
    artist = ";laksjdflkajsdfj;asdjf"
    song = "jaksjd;fljkas;dfkjasldkjf"
    result = music_yt.search(artist, song)
    assert result is None


def test_search_partial_match(music_yt):
    artist = "Adele"
    song = "Hello"
    result = music_yt.search(artist, song)
    assert result is not None
    assert "Hello" in result.title  # Assuming the title contains the song name


def test_search_empty_artist(music_yt):
    artist = ""
    song = "Hello"

    with raises(InvalidValueException):
        music_yt.search(artist, song)


def test_search_empty_song(music_yt):
    artist = "Adele"
    song = ""

    with raises(InvalidValueException):
        music_yt.search(artist, song)


def test_close_session(music_yt):
    music_yt.close_session()
    assert music_yt.is_session_closed



__all__ = ["KKBox", "KKBoxException"]

import base64
import os
import time
from pprint import pprint
from typing import Optional, Union

import requests
from dotenv import load_dotenv

from yutipy.exceptions import (
    AuthenticationException,
    InvalidResponseException,
    InvalidValueException,
    KKBoxException,
    NetworkException,
)
from yutipy.models import MusicInfo
from yutipy.utils.helpers import are_strings_similar, is_valid_string
from yutipy.utils.logger import logger

load_dotenv()

KKBOX_CLIENT_ID = os.getenv("KKBOX_CLIENT_ID")
KKBOX_CLIENT_SECRET = os.getenv("KKBOX_CLIENT_SECRET")


class KKBox:
    """
    A class to interact with KKBOX Open API.

    This class reads the ``KKBOX_CLIENT_ID`` and ``KKBOX_CLIENT_SECRET`` from environment variables or the ``.env`` file by default.
    Alternatively, you can manually provide these values when creating an object.
    """

    def __init__(
        self, client_id: str = KKBOX_CLIENT_ID, client_secret: str = KKBOX_CLIENT_SECRET
    ) -> None:
        """
        Initializes the KKBox class and sets up the session.

        Parameters
        ----------
        client_id : str, optional
            The Client ID for the KKBOX Open API. Defaults to ``KKBOX_CLIENT_ID`` from .env file.
        client_secret : str, optional
            The Client secret for the KKBOX Open API. Defaults to ``KKBOX_CLIENT_SECRET`` from .env file.
        """
        if not client_id or not client_secret:
            raise KKBoxException(
                "Failed to read `KKBOX_CLIENT_ID` and/or `KKBOX_CLIENT_SECRET` from environment variables. Client ID and Client Secret must be provided."
            )

        self.client_id = client_id
        self.client_secret = client_secret
        self._session = requests.Session()
        self.api_url = "https://api.kkbox.com/v1.1"
        self.__header, self.__expires_in = self.__authenticate()
        self.__start_time = time.time()
        self._is_session_closed = False
        self.valid_territories = ["HK", "JP", "MY", "SG", "TW"]
        self.normalize_non_english = True
        self._translation_session = requests.Session()

    def __enter__(self):
        """Enters the runtime context related to this object."""
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        """Exits the runtime context related to this object."""
        self.close_session()

    def close_session(self) -> None:
        """Closes the current session."""
        if not self.is_session_closed:
            self._session.close()
            self._translation_session.close()
            self._is_session_closed = True

    @property
    def is_session_closed(self) -> bool:
        """Checks if the session is closed."""
        return self._is_session_closed

    def __authenticate(self) -> tuple:
        """
        Authenticates with the KKBOX Open API and returns the authorization header.

        Returns
        -------
        dict
            The authorization header.
        """
        try:
            token, expires_in = self.__get_access_token()
            return {"Authorization": f"Bearer {token}"}, expires_in
        except Exception as e:
            raise AuthenticationException(
                "Failed to authenticate with KKBOX Open API"
            ) from e

    def __get_access_token(self) -> tuple:
        """
        Gets the KKBOX Open API token.

        Returns
        -------
        str
            The KKBOX Open API token.
        """
        auth_string = f"{self.client_id}:{self.client_secret}"
        auth_base64 = base64.b64encode(auth_string.encode("utf-8")).decode("utf-8")

        url = " https://account.kkbox.com/oauth2/token"
        headers = {
            "Authorization": f"Basic {auth_base64}",
            "Content-Type": "application/x-www-form-urlencoded",
        }
        data = {"grant_type": "client_credentials"}

        try:
            logger.info("Authenticating with KKBOX Open API")
            response = self._session.post(
                url=url, headers=headers, data=data, timeout=30
            )
            logger.debug(f"Authentication response status code: {response.status_code}")
            response.raise_for_status()
        except requests.RequestException as e:
            logger.error(f"Network error during KKBOX authentication: {e}")
            raise NetworkException(f"Network error occurred: {e}")

        try:
            response_json = response.json()
            return response_json.get("access_token"), response_json.get("expires_in")
        except (KeyError, ValueError) as e:
            raise InvalidResponseException(f"Invalid response received: {e}")

    def __refresh_token_if_expired(self):
        """Refreshes the token if it has expired."""
        if time.time() - self.__start_time >= self.__expires_in:
            self.__header, self.__expires_in = self.__authenticate()
            self.__start_time = time.time()

    def search(
        self,
        artist: str,
        song: str,
        territory: str = "TW",
        limit: int = 10,
        normalize_non_english: bool = True,
    ) -> Optional[MusicInfo]:
        """
        Searches for a song by artist and title.

        Parameters
        ----------
        artist : str
            The name of the artist.
        song : str
            The title of the song.
        territory : str
            Two-letter country codes from ISO 3166-1 alpha-2.
            Allowed values: ``HK``, ``JP``, ``MY``, ``SG``, ``TW``.
        limit: int, optional
            The number of items to retrieve from API. ``limit >=1 and <= 50``. Default is ``10``.
        normalize_non_english : bool, optional
            Whether to normalize non-English characters for comparison. Default is ``True``.

        Returns
        -------
        Optional[MusicInfo_]
            The music information if found, otherwise None.
        """
        if not is_valid_string(artist) or not is_valid_string(song):
            raise InvalidValueException(
                "Artist and song names must be valid strings and can't be empty."
            )

        self.normalize_non_english = normalize_non_english

        self.__refresh_token_if_expired()

        query = (
            f"?q={artist} - {song}&type=track,album&territory={territory}&limit={limit}"
        )
        query_url = f"{self.api_url}/search{query}"

        logger.info(f"Searching KKBOX for `artist='{artist}'` and `song='{song}'`")
        logger.debug(f"Query URL: {query_url}")

        try:
            response = self._session.get(query_url, headers=self.__header, timeout=30)
            logger.debug(f"Parsing response JSON: {response.json()}")
            response.raise_for_status()
        except requests.RequestException as e:
            raise NetworkException(f"Network error occurred: {e}")

        if response.status_code != 200:
            raise KKBoxException(f"Failed to search for music: {response.json()}")

        return self._find_music_info(artist, song, response.json())

    def get_html_widget(
        self,
        id: str,
        content_type: str,
        territory: str = "TW",
        widget_lang: str = "EN",
        autoplay: bool = False,
        loop: bool = False,
    ) -> str:
        """
        Return KKBOX HTML widget for "Playlist", "Album" or "Song". It does not return actual HTML code,
        the URL returned can be used in an HTML ``iframe`` with the help of ``src`` attribute.

        Parameters
        ----------
        id : str
             ``ID`` of playlist, album or track.
        content_type : str
            Content type can be ``playlist``, ``album`` or ``song``.
        territory : str, optional
            Territory code, i.e. "TW", "HK", "JP", "SG", "MY", by default "TW"
        widget_lang : str, optional
            The display language of the widget. Can be "TC", "SC", "JA", "EN", "MS", by default "EN"
        autoplay : bool, optional
            Whether to start playing music automatically in widget, by default False
        loop : bool, optional
            Repeat/loop song(s), by default False

        Returns
        -------
        str
            KKBOX HTML widget URL.
        """
        valid_content_types = ["playlist", "album", "song"]
        valid_widget_langs = ["TC", "SC", "JA", "EN", "MS"]
        if content_type not in valid_content_types:
            raise InvalidValueException(
                f"`content_type` must be one of these: {valid_content_types} !"
            )

        if territory not in self.valid_territories:
            raise InvalidValueException(
                f"`territory` must be one of these: {self.valid_territories} !"
            )

        if widget_lang not in valid_widget_langs:
            raise InvalidValueException(
                f"`widget_lang` must be one of these: {valid_widget_langs} !"
            )

        return f"https://widget.kkbox.com/v1/?id={id}&type={content_type}&terr={territory}&lang={widget_lang}&autoplay={autoplay}&loop={loop}"

    def _find_music_info(
        self, artist: str, song: str, response_json: dict
    ) -> Optional[MusicInfo]:
        """
        Finds the music information from the search results.

        Parameters
        ----------
        artist : str
            The name of the artist.
        song : str
            The title of the song.
        response_json : dict
            The JSON response from the API.

        Returns
        -------
        Optional[MusicInfo]
            The music information if found, otherwise None.
        """
        try:
            for track in response_json["tracks"]["data"]:
                music_info = self._find_track(song, artist, track)
                if music_info:
                    return music_info
        except KeyError:
            pass

        try:
            for album in response_json["albums"]["data"]:
                music_info = self._find_album(song, artist, album)
                if music_info:
                    return music_info
        except KeyError:
            pass

        logger.warning(
            f"No matching results found for artist='{artist}' and song='{song}'"
        )
        return None

    def _find_track(self, song: str, artist: str, track: dict) -> Optional[MusicInfo]:
        """
        Finds the track information from the search results.

        Parameters
        ----------
        song : str
            The title of the song.
        artist : str
            The name of the artist.
        track : dict
            A single track from the search results.
        artist_ids : list
            A list of artist IDs.

        Returns
        -------
        Optional[MusicInfo]
            The music information if found, otherwise None.
        """
        if not are_strings_similar(
            track["name"],
            song,
            use_translation=self.normalize_non_english,
            translation_session=self._translation_session,
        ):
            return None

        artists_name = track["album"]["artist"]["name"]
        matching_artists = (
            artists_name
            if are_strings_similar(
                artists_name,
                artist,
                use_translation=self.normalize_non_english,
                translation_session=self._translation_session,
            )
            else None
        )

        if matching_artists:
            return MusicInfo(
                album_art=track["album"]["images"][2]["url"],
                album_title=track["album"]["name"],
                album_type=None,
                artists=artists_name,
                genre=None,
                id=track["id"],
                isrc=track["isrc"],
                lyrics=None,
                release_date=track["album"]["release_date"],
                tempo=None,
                title=track["name"],
                type="track",
                upc=None,
                url=track["url"],
            )

        return None

    def _find_album(self, song: str, artist: str, album: dict) -> Optional[MusicInfo]:
        """
        Finds the album information from the search results.

        Parameters
        ----------
        song : str
            The title of the song.
        artist : str
            The name of the artist.
        album : dict
            A single album from the search results.
        artist_ids : list
            A list of artist IDs.

        Returns
        -------
        Optional[MusicInfo]
            The music information if found, otherwise None.
        """
        if not are_strings_similar(
            album["name"],
            song,
            use_translation=self.normalize_non_english,
            translation_session=self._translation_session,
        ):
            return None

        artists_name = album["artist"]["name"]
        matching_artists = (
            artists_name
            if are_strings_similar(
                artists_name,
                artist,
                use_translation=self.normalize_non_english,
                translation_session=self._translation_session,
            )
            else None
        )

        if matching_artists:
            return MusicInfo(
                album_art=album["images"][2]["url"],
                album_title=album["name"],
                album_type=None,
                artists=artists_name,
                genre=None,
                id=album["id"],
                isrc=None,
                lyrics=None,
                release_date=album["release_date"],
                tempo=None,
                title=album["name"],
                type="album",
                upc=None,
                url=album["url"],
            )

        return None


if __name__ == "__main__":
    import logging
    from yutipy.utils.logger import enable_logging

    enable_logging(level=logging.DEBUG)
    kkbox = KKBox(KKBOX_CLIENT_ID, KKBOX_CLIENT_SECRET)

    try:
        artist_name = input("Artist Name: ")
        song_name = input("Song Name: ")
        pprint(kkbox.search(artist_name, song_name))
    finally:
        kkbox.close_session()
